"""tracker test application"""

from datetime import date

from cubicweb.devtools.testlib import CubicWebTC, AutomaticWebTest


class AutomaticWebTest(AutomaticWebTest):

    def to_test_etypes(self):
        return set(('Project', 'Ticket', 'Version'))

    def list_startup_views(self):
        return ('stats', 'versionsinfo')

    def post_populate(self, cursor):
        self.commit()
        for version in cursor.execute('Version X').entities():
            version.change_state('published')


class VersionCalendarViews(CubicWebTC):
    """specific tests for calendar views"""

    def setup_database(self):
        req = self.request()
        req.create_entity('Version', num=u'0.1.0', publication_date=date(2006, 2, 1))
        req.create_entity('Version', num=u'0.2.0', publication_date=date(2006, 4, 1))
        req.create_entity('Project', name=u"MyProject")
        req.execute('SET V version_of P where V is Version, P is Project')

    def test_calendars_for_versions(self):
        rset = self.execute('Version V')
        for vid in ('onemonthcal', 'oneweekcal'):
            yield self.view, vid, rset, rset.req.reset_headers()

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



"""some utilities to define tracker schema permissions

:organization: Logilab
:copyright: 2008 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from rql.utils import quote
from cubicweb.schema import ERQLExpression, RRQLExpression

def _perm(names):
    if isinstance(names, (list, tuple)):
        if len(names) == 1:
            names = quote(names[0])
        else:
            names = 'IN (%s)' % (','.join(quote(name) for name in names))
    else:
        names = quote(names)
    #return u' require_permission P, P name %s, U in_group G, P require_group G' % names
    return u' require_permission P, P name %s, U has_group_permission P' % names


def xperm(*names):
    return 'X' + _perm(names)

def xexpr(*names):
    return ERQLExpression(xperm(*names))

def xrexpr(relation, *names):
    return ERQLExpression('X %s Y, Y %s' % (relation, _perm(names)))

def xorexpr(relation, etype, *names):
    return ERQLExpression('Y %s X, X is %s, Y %s' % (relation, etype, _perm(names)))


def sexpr(*names):
    return RRQLExpression('S' + _perm(names), 'S')

def restricted_sexpr(restriction, *names):
    rql = '%s, %s' % (restriction, 'S' + _perm(names))
    return RRQLExpression(rql, 'S')

def restricted_oexpr(restriction, *names):
    rql = '%s, %s' % (restriction, 'O' + _perm(names))
    return RRQLExpression(rql, 'O')

def oexpr(*names):
    return RRQLExpression('O' + _perm(names), 'O')


# def supdate_perm():
#     return RRQLExpression('U has_update_permission S', 'S')

# def oupdate_perm():
#     return RRQLExpression('U has_update_permission O', 'O')
def relxperm(rel, role, *names):
    assert role in ('subject', 'object')
    if role == 'subject':
        zxrel = ', X %s Z' % rel
    else:
        zxrel = ', Z %s X' % rel
    return 'Z' + _perm(names) + zxrel

def relxexpr(rel, role, *names):
    return ERQLExpression(relxperm(rel, role, *names))



"""tracker specific entities class for imported entities

:organization: Logilab
:copyright: 2006-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

def fixed_orderby_rql(sortsdef, asc=True):
    orderby = []
    for rtype, varname in sortsdef:
        if rtype == 'priority':
            orderby.append('priority_sort_value(%s)' % varname)
        elif rtype == 'num':
            orderby.append('version_sort_value(%s)' % varname)
        else:
            orderby.append(varname)
    if asc:
        return 'ORDERBY %s' % ', '.join(orderby)
    return 'ORDERBY %s DESC' % ', '.join(orderby)


# library overrides ###########################################################

class ProjectItemMixIn(object):
    """default mixin class for commentable objects to make them implement
    the project item interface. Defined as a mixin for use by custom tracker
    templates.
    """
    @property
    def project(self):
        # XXX should we try to pick up one of the see_also projects ?
        # rset = self.req.execute('Any P WHERE P is Project, X see_also P, X eid %(x)s LIMIT 1',
        #                         {'x': self.eid})
        return None

    def after_deletion_path(self):
        """return (path, parameters) which should be used as redirect
        information when this entity is being deleted
        """
        if self.project:
            return self.project.rest_path(), {}
        return 'view', {}

from cubicweb.schema import SYSTEM_RTYPES
SYSTEM_RTYPES.add('has_group_permission')



"""views for Ticket entities

:organization: Logilab
:copyright: 2006-2009 LOGILAB S.A. (Paris, FRANCE), license is LGPL.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from logilab.mtconverter import xml_escape

from cubicweb.view import EntityView
from cubicweb.schema import display_name
from cubicweb.selectors import (objectify_selector, implements, multi_lines_rset,
                                score_entity)
from cubicweb import tags, uilib
from cubicweb.web import uicfg, component, action
from cubicweb.web.views import primary, baseviews

_pvs = uicfg.primaryview_section

_pvs.tag_attribute(('Ticket', 'title'), 'hidden')
_pvs.tag_attribute(('Ticket', 'description'), 'attributes')

_pvs.tag_subject_of(('Ticket', 'concerns', 'Project'), 'hidden')
_pvs.tag_object_of(('Ticket', 'concerns', 'Project'), 'hidden')
_pvs.tag_subject_of(('Ticket', 'done_in', 'Version'), 'attributes')
_pvs.tag_object_of(('Ticket', 'done_in', 'Version'), 'attributes')

_pvs.tag_subject_of(('Ticket', 'appeared_in', '*'), 'sideboxes')
_pvs.tag_object_of(('Ticket', 'appeared_in', '*'), 'hidden')
_pvs.tag_subject_of(('Ticket', 'depends_on', '*'), 'sideboxes')
_pvs.tag_object_of(('*', 'depends_on', 'Ticket'), 'sideboxes')

_pvdc = uicfg.primaryview_display_ctrl
_pvdc.tag_attribute(('Ticket', 'description'), {'showlabel': False})

# primary view and tabs ########################################################

class TicketPrimaryView(primary.PrimaryView):
    """primary view for tickets
    """
    __select__ = primary.PrimaryView.__select__ & implements('Ticket')

    def render_entity_title(self, entity):
        self.w(u'<h1><span class="%s %s">%s %s</span><span class="state"> [%s]</span></h1>'
               % (entity.priority, entity.type,
                  xml_escape(entity.project.name),
                  xml_escape(entity.dc_title()),
                  xml_escape(self._cw._(entity.state))))


# pluggable sections ###########################################################

class TicketIdenticalToVComponent(component.RelatedObjectsVComponent):
    """display identical tickets"""
    __regid__ = 'tickectidentical'
    __select__ = component.RelatedObjectsVComponent.__select__ & implements('Ticket')

    rtype = 'identical_to'
    target = 'object'

    title = _('Identical tickets')
    context = 'navcontentbottom'
    order = 20


# secondary views ##############################################################

class TicketOneLineView(baseviews.OneLineView):
    """one line representation of a ticket:

    call text view to get link's label, set entity's description on the
    link and display entity's status
    """
    __select__ = implements('Ticket')

    def cell_call(self, row, col):
        self.wview('incontext', self.cw_rset, row=row)
        entity = self.cw_rset.get_entity(row, col)
        if entity.in_state:
            self.w(u'&nbsp;[%s]' % xml_escape(self._cw._(entity.state)))


class TicketInContextView(baseviews.OneLineView):
    """in-context representation of a ticket:

    call text view to get link's label, set entity's description on the
    link and display entity's status
    """
    __regid__ = 'incontext'
    __select__ = implements('Ticket')

    def cell_call(self, row, col):
        self._cw.add_css('cubes.tracker.css')
        entity = self.cw_rset.get_entity(row, col)
        self.w(tags.a(entity.dc_title(), href=entity.absolute_url(),
                      title=uilib.cut(entity.dc_description(), 80),
                      klass=entity.priority))


class StatusSheetTicketView(EntityView):
    __regid__ = 'instatussheet'
    __select__ = implements('Ticket')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(tags.div(tags.a('T%s' % entity.eid,
                               href=xml_escape(entity.absolute_url())),
                        title=xml_escape(entity.title),
                        style='display: inline'))


# actions ######################################################################

@objectify_selector
def ticket_has_next_version(cls, req, rset, row=None, col=0, **kwargs):
    rschema = req.vreg.schema.rschema('done_in')
    if row is None:
        # action is applyable if all entities are ticket from the same project,
        # in an open state, share some versions to which they may be moved
        project, versions = None, set()
        for entity in rset.entities():
            if entity.e_schema != 'Ticket':
                return 0
            if not entity.is_open():
                return 0
            if project is None:
                project = entity.project
            elif project.eid != entity.project.eid:
                return 0
            if entity.in_version():
                versions.add(entity.in_version().eid)
        if project is None:
            return 0
        maymoveto = []
        for version in project.versions_in_state(('planned', 'dev')).entities():
            if version.eid in versions:
                continue
            for entity in rset.entities():
                if not rschema.has_perm(req, 'add', fromeid=entity.eid,
                                        toeid=version.eid):
                    break
            else:
                maymoveto.append(version)
        if maymoveto:
            rset.maymovetoversions = maymoveto # cache for use in action
            return 1
        return 0
    entity = rset.get_entity(row, 0)
    versionsrset = entity.project.versions_in_state(('planned', 'dev'))
    if not versionsrset:
        return 0
    ticketversion = entity.in_version() and entity.in_version().eid
    maymoveto = [version for version in versionsrset.entities()
                 if not version.eid == ticketversion and
                 rschema.has_perm(req, 'add', fromeid=entity.eid,
                                  toeid=version.eid)]
    if maymoveto:
        rset.maymovetoversions = maymoveto # cache for use in action
        return 1
    return 0


class TicketAction(action.Action):
    __select__ = action.Action.__select__ & implements('Ticket')
    # use "mainactions" category to appears in table filter's actions menu
    category = 'mainactions'


class TicketMoveToNextVersionActions(TicketAction):
    __regid__ = 'movetonext'
    __select__ = (TicketAction.__select__
                  & score_entity(lambda x: x.state in x.OPEN_STATES)
                  & ticket_has_next_version())

    submenu = _('move to version')

    def fill_menu(self, box, menu):
        # when there is only one item in the sub-menu, replace the sub-menu by
        # item's title prefixed by 'move to version'
        menu.label_prefix = self._cw._(self.submenu)
        super(TicketMoveToNextVersionActions, self).fill_menu(box, menu)

    def actual_actions(self):
        for version in self.cw_rset.maymovetoversions:
            yield self.build_action(version.num, self.url(version))

    def url(self, version):
        if self.cw_row is None:
            eids = [str(row[self.cw_col or 0]) for row in self.cw_rset]
        else:
            eids = [str(self.cw_rset[self.cw_row][self.cw_col or 0])]
        rql = 'SET X done_in V WHERE X eid IN(%s), V eid %%(v)s' % ','.join(eids)
        msg = self._cw._('tickets moved to version %s') % version.num
        return self._cw.user_rql_callback((rql, {'v': version.eid}, 'v'), msg)


class TicketCSVExportAction(TicketAction):
    __regid__ = 'ticketcsvexport'
    __select__ = multi_lines_rset() & TicketAction.__select__

    title = _('csv export')

    def url(self):
        return self._cw.build_url('view', rql=self.cw_rset.printable_rql(),
                              vid='csvexport')



"""tracker web user interface

:organization: Logilab
:copyright: 2006-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from cubicweb.web import uicfg
from cubicweb.web.views import schema
from cubicweb.web.views.urlrewrite import (
    SimpleReqRewriter, SchemaBasedRewriter, rgx, build_rset)


schema.SKIP_TYPES.add('see_also')
schema.ALWAYS_SKIP_TYPES.add('see_also')


class TrackerSimpleReqRewriter(SimpleReqRewriter):
    """handle path with the form::

    versions -> view versions in state ready or development, or marked as
                prioritary.

    foaf.rdf -> view the foaf file for all users
    """
    rules = [
        (rgx('/versions'), dict(vid='versionsinfo')),
        ('/foaf.rdf', dict(vid='foaf', rql='Any X WHERE X is CWUser'))
        ]

class TrackerURLRewriter(SchemaBasedRewriter):
    """handle path with the form::

        project/<name>/tickets           -> view project's tickets
        project/<name>/versions          -> view project's versions in state ready
                                            or development, or marked as
                                            prioritary.
        project/<name>/[version]         -> view version
        project/<name>/[version]/tests   -> test for this version
        project/<name>/[version]/tickets -> tickets for this version
    """
    rules = [
        (rgx('/project/([^/]+)/([^/]+)/tests'),
         build_rset(rql='Version X WHERE X version_of P, P name %(project)s, X num %(num)s',
                    rgxgroups=[('project', 1), ('num', 2)], vid='versiontests')),
        (rgx('/project/([^/]+)/([^/]+)/tickets'),
         build_rset(rql='Any T WHERE T is Ticket, T done_in V, V version_of P, P name %(project)s, V num %(num)s',
                    rgxgroups=[('project', 1), ('num', 2)],
                    vtitle=_('tickets for %(project)s - %(num)s'))),
        (rgx('/project/([^/]+)/versions'),
         build_rset(rql='Any X,N ORDERBY version_sort_value(N) '
                    'WHERE X num N, X version_of P, P name %(project)s, '
                    'EXISTS(X in_state S, S name IN ("dev", "ready")) ',
                    rgxgroups=[('project', 1)], vid='ic_progress_table_view',
                    vtitle=_('upcoming versions for %(project)s'))),
        (rgx('/project/([^/]+)/tickets'),
         build_rset(rql='Any T WHERE T is Ticket, T concerns P, P name %(project)s',
                    rgxgroups=[('project', 1)], vid='table',
                    vtitle=_('tickets for %(project)s'))),
        (rgx('/project/([^/]+)/([^/]+)'),
         build_rset(rql='Version X WHERE X version_of P, P name %(project)s, X num %(num)s',
                    rgxgroups=[('project', 1), ('num', 2)])),
        (rgx('/p/([^/]+)'),
         build_rset(rql='Project P WHERE P name %(project)s',
                    rgxgroups=[('project', 1),])),
        (rgx('/t/([^/]+)'),
         build_rset(rql='Ticket T WHERE T eid %(teid)s',
                    rgxgroups=[('teid', 1),])),
         ]


_pvs = uicfg.primaryview_section
_pvs.tag_object_of(('*', 'todo_by', 'CWUser'), 'relations')

uicfg.primaryview_display_ctrl.tag_object_of(
    ('*', 'todo_by', 'CWUser'),
    {'vid': 'list', 'label': _('working on release(s):'), 'limit': False,
     'filter': lambda rset: rset.filtered_rset(lambda x: x.state == u'dev'),
     })

_abaa = uicfg.actionbox_appearsin_addmenu
_abaa.tag_subject_of(('*', 'todo_by', '*'), False)



from cubicweb.selectors import implements
from cubicweb.web.facet import RelationFacet, AttributeFacet

class TicketConcernsFacet(RelationFacet):
    __regid__ = 'concerns-facet'
    __select__ = RelationFacet.__select__ & implements('Ticket')
    rtype = 'concerns'
    target_attr = 'name'


class TicketDoneInFacet(RelationFacet):
    __regid__ = 'done_in-facet'
    __select__ = RelationFacet.__select__ & implements('Ticket')
    rtype = 'done_in'
    target_attr = 'num'
    sortfunc = 'VERSION_SORT_VALUE'
    sortasc = False


class TicketPriorityFacet(AttributeFacet):
    __regid__ = 'priority-facet'
    __select__ = AttributeFacet.__select__ & implements('Ticket')
    rtype = 'priority'
    sortfunc = 'PRIORITY_SORT_VALUE'


class TicketTypeFacet(AttributeFacet):
    __regid__ = 'type-facet'
    __select__ = AttributeFacet.__select__ & implements('Ticket')
    rtype = 'type'



# register priority/severity sorting stored procedures
from rql.utils import register_function, FunctionDescr


class priority_sort_value(FunctionDescr):
    supported_backends = ('postgres', 'sqlite',)
    rtype = 'Int'
class version_sort_value(FunctionDescr):
    supported_backends = ('postgres', 'sqlite',)
    rtype = 'Int'

try:
    register_function(priority_sort_value)
    register_function(version_sort_value)
except AssertionError:
    pass



try:
    from cubicweb.server import SQL_CONNECT_HOOKS
except ImportError: # no server installation
    pass
else:

    def init_sqlite_connexion(cnx):
        def priority_sort_value(text):
            return {"minor": "2", "normal": "1", "important":  "0"}[text]
        cnx.create_function("PRIORITY_SORT_VALUE", 1, priority_sort_value)

        def version_sort_value(text):
            try:
                return sum((10000, 100, 1)[i]*int(v) for i, v in enumerate(text.split(".")))
            except (ValueError, IndexError, AttributeError):
                return 0
        cnx.create_function("VERSION_SORT_VALUE", 1, version_sort_value)

    sqlite_hooks = SQL_CONNECT_HOOKS.setdefault('sqlite', [])
    sqlite_hooks.append(init_sqlite_connexion)

# XML <-> yams equivalence
from cubicweb.xy import xy
xy.add_equivalence('Project', 'doap:Project')
xy.add_equivalence('Project creation_date', 'doap:Project doap:created')



"""tracker server side objects, mainly notification stuff

:organization: Logilab
:copyright: 2006-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"


from logilab.common.textutils import normalize_text

from cubicweb.selectors import implements, score_entity
from cubicweb.mail import SkipEmail
from cubicweb.sobjects.notification import (StatusChangeMixIn, NotificationView,
                                            EntityUpdatedNotificationView)

class TrackerEmailView(NotificationView):

    def subject(self):
        entity = self.cw_rset.get_entity(0,0)
        return '[%s] %s' % (entity.project.name, self._subject(entity))

    def _subject(self, entity):
        return '%s %s (%s)' % (entity.dc_type(), self.message, entity.dc_title())


class TicketPropertiesChangeView(TrackerEmailView, EntityUpdatedNotificationView):
    no_detailed_change_attrs =  ('description', 'description_format')
    __select__ = implements('Ticket')

    content = _("""
Ticket properties have been updated by %(user)s:
%(changes)s

url: %(url)s
""")

    def _subject(self, entity):
        return self._cw._(u'%(etype)s "%(title)s" updated') % {
            'etype': entity.dc_type(), 'title': entity.dc_title()}


class VersionChangedView(TicketPropertiesChangeView):
    __select__ = implements('Version')
    content = _("""
Version dates have been updated by %(user)s:
%(changes)s

url: %(url)s
""")


class VersionStatusChangeView(StatusChangeMixIn, TrackerEmailView):
    __select__ = implements('Version')

    def _subject(self, entity):
        return self._cw._(u'version %(num)s is now in state "%(state)s"') % {
            'num': entity.num,
            'state': self._cw.__(self._kwargs['current_state'])}


class VersionPublishedView(VersionStatusChangeView):
    __select__ = (VersionStatusChangeView.__select__
                  # use latest trinfo since we're not yet linked to the new
                  # state at this point
                  & score_entity(lambda x: x.latest_trinfo().new_state.name in x.PUBLISHED_STATES)
                  )

    content = _("""
Bugs fixed in this release:
\t%(bugslist)s

Enhancements implemented in this release:
\t%(enhancementslist)s

Tasks done in this release:
\t%(taskslist)s

url: %(url)s
""")

    def context(self, **kwargs):
        context = super(VersionPublishedView, self).context(**kwargs)
        entity = self.cw_rset.get_entity(0, 0)
        bugs = ['- '+t.dc_title() for t in entity.reverse_done_in if t.type == 'bug']
        enhancements = ['- '+t.dc_title() for t in entity.reverse_done_in if t.type == 'enhancement']
        tasks = ['- '+t.dc_title() for t in entity.reverse_done_in if t.type == 'task']
        context['bugslist'] = '\n\t'.join(bugs)
        context['enhancementslist'] = '\n\t'.join(enhancements)
        context['taskslist'] = '\n\t'.join(tasks)
        return context


class TicketStatusChangeView(StatusChangeMixIn, TrackerEmailView):
    __select__ = implements('Ticket')

    def _subject(self, entity):
        return self._cw._(u'%(etype)s "%(title)s" is now in state "%(state)s"') % {
            'etype': entity.dc_type(), 'title': entity.dc_title(),
            'state': self._cw.__(self._kwargs['current_state'])}


class ProjectAddedView(TrackerEmailView):
    __regid__ = 'notif_after_add_entity'
    __select__ = implements('Project')

    section_attrs = ['summary', 'description']
    content = _("""
A new project was created by %(user)s: #%(eid)s - %(pname)s

%(pcontent)s

URL
---
%(url)s
""")

    def context(self, **kwargs):
        context = super(ProjectAddedView, self).context(**kwargs)
        entity = self.cw_rset.get_entity(0, 0)
        sections = []
        for attr in self.section_attrs:
            val = entity.printable_value(attr, format='text/plain')
            if val:
                sect = self.format_section(self._cw._(attr).capitalize(), val)
                sections.append(sect)
        context['pcontent'] = '\n'.join(sections)
        context['pname'] = entity.name
        return context

    def subject(self):
        return u'[%s] %s: %s' % (
            self._cw.vreg.config.appid, self._cw.__('New Project'), self.cw_rset.get_entity(0,0).name)


class TicketSubmittedView(TrackerEmailView):
    __regid__ = 'notif_after_add_relation_concerns'
    __select__ = implements('Ticket')
    content = _("""
New %(etype)s for project %(pname)s :

#%(eid)s - %(title)s
====================
%(mainsection)s

description
-----------
%(description)s

submitter
---------
%(user)s

URL
---
%(url)s
(project URL: %(purl)s)
""")
    field_attrs = ['type', 'priority']

    def context(self, **kwargs):
        ctx = super(TicketSubmittedView, self).context(**kwargs)
        entity = self.cw_rset.get_entity(0, 0)
        sect = []
        for attr in self.field_attrs:
            sect.append(self.format_field(attr, entity.printable_value(attr)))
        ctx['mainsection'] = '\n'.join(sect)
        description = entity.printable_value('description', format='text/plain')
        description = normalize_text(description, 80)
        ctx['description'] = description
        ctx['pname'] = entity.project.name
        ctx['purl'] = entity.project.absolute_url()
        return ctx

    def _subject(self, entity):
        return '%s %s' % (self._cw.__('New %s' % entity.e_schema),
                          entity.dc_title())



for v in rql('Any V, VP WHERE V is Version, '
             'V publication_date VP, V publication_date NULL, '
             'V in_state S, S name "published"').entities():
    v.set_attributes(publication_date=v.latest_trinfo().creation_date)
    print 'fixed publication date of', v.dc_long_title()
commit()



"""tracker specific entities class for Ticket

:organization: Logilab
:copyright: 2006-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

__docformat__ = "restructuredtext en"

from cubicweb.interfaces import IPrevNext
from cubicweb.entities import AnyEntity


class Ticket(AnyEntity):
    __regid__ = 'Ticket'
    __permissions__ = ('developer', 'client')
    __implements__ = AnyEntity.__implements__ + (IPrevNext,)
    fetch_attrs = ('title', 'type', 'priority', 'in_state')

    OPEN_STATES = frozenset(('in-progress',))

    @classmethod
    def fetch_order(cls, attr, var):
        if attr == 'priority':
            return 'priority_sort_value(%s)' % var
        if attr in ('title', 'type'):
            return var
        return None

    @property
    def project(self):
        """project item interface"""
        return self.concerns[0]

    # hierarchy #############################################################

    def parent(self):
        parents = self.done_in or self.concerns
        if parents:
            return parents[0]

    # dublin core #############################################################

    def dc_title(self):
        return u'#%s: %s' % (self.eid, self.title)

    def dc_long_title(self):
        return u'%s %s' % (self.project.name, self.dc_title())

    # ticket'specific logic ###################################################


    def is_open(self):
        return self.state in self.OPEN_STATES

    def state_history(self):
        """returns a list of dates where the state changed from "open" to "not-open"
        :returns: a list couples of (date, is_open)
        """
        dates = [(self.creation_date, True)]
        for tr_info in self.reverse_wf_info_for:
            prevstate = tr_info.previous_state.name
            nextstate = tr_info.new_state.name
            if prevstate in self.OPEN_STATES and nextstate not in self.OPEN_STATES:
                dates.append( (tr_info.creation_date, False) )
            elif prevstate not in self.OPEN_STATES and nextstate in self.OPEN_STATES:
                dates.append( (tr_info.creation_date, True) )
        return sorted(dates)

    def in_version(self):
        versions = self.done_in
        if versions:
            return versions[0]

    def next_version(self):
        version = self.in_version()
        if version:
            return version.next_version()
        return self.project.versions_in_state(('planned', 'dev'), True)

    def sortvalue(self, rtype=None):
        if rtype == 'priority':
            return ['minor', 'normal', 'important'].index(self.priority)
        return super(Ticket, self).sortvalue(rtype)

    def after_deletion_path(self):
        """return (path, parameters) which should be used as redirect
        information when this entity is being deleted
        """
        if self.done_in:
            return self.done_in[0].rest_path(), {}
        return self.project.rest_path(), {}

    # IPrevNext interface #######################################

    def previous_entity(self):
        rql = ('Any X,T ORDERBY X DESC LIMIT 1 '
               'WHERE X is Ticket, X concerns P, X title T, P eid %(p)s, '
               'X eid < %(x)s')
        rset = self._cw.execute(rql, {'p': self.project.eid, 'x': self.eid})
        if rset:
            return rset.get_entity(0, 0)

    def next_entity(self):
        rql = ('Any X,T ORDERBY X ASC LIMIT 1 '
               'WHERE X is Ticket, X concerns P, X title T, P eid %(p)s, '
               'X eid > %(x)s')
        rset = self._cw.execute(rql, {'p': self.project.eid, 'x': self.eid})
        if rset:
            return rset.get_entity(0, 0)



# pylint:  disable-msg=E0611,F0401
"""tracker application'schema

:organization: Logilab
:copyright: 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from yams.constraints import SizeConstraint
from yams.buildobjs import (
    EntityType, RelationType, RelationDefinition,
    SubjectRelation, ObjectRelation, RichString, String, Date)

from cubicweb.schema import (
    RQLConstraint, RQLVocabularyConstraint,
    ERQLExpression, RRQLExpression, WorkflowableEntityType)

from cubes.tracker.schemaperms import (
    sexpr, oexpr, xexpr, restricted_oexpr, xperm)


class Project(EntityType):
    """a project is a logical group of tasks/information which produce a final value. That
    may be a software or documentation project, a web site or whatever.
    """
    __permissions__ = {
        'read':   ('managers', 'users', 'guests',),
        'add':    ('managers', 'staff',),
        'update': ('managers', 'staff', 'owners',),
        'delete': ('managers', 'owners'),
        }
    ## attributes
    name      = String(required=True, fulltextindexed=True, unique=True, maxsize=64)
    summary   = String(fulltextindexed=True, maxsize=128,
                       description=_('one line description of the project'))

    description = RichString(fulltextindexed=True, default_format='text/rest',
                             description=_('more detailed description'))

    ## relations
    uses = SubjectRelation('Project', description=_('project\'s dependencies'))
    subproject_of = SubjectRelation('Project', composite='object', cardinality='?*')


class uses(RelationType):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers', RRQLExpression('U has_update_permission S', 'S'),),
        'delete': ('managers', RRQLExpression('U has_update_permission S', 'S'),),
        }


class Version(WorkflowableEntityType):
    """a version is defining the content of a particular project's release"""
    __permissions__ = {
        'read':   ('managers', 'users', 'guests',),
        # right to add Version is actually granted on the version_of relation
        'add':    ('managers', 'users'),
        'update': ('managers', 'staff', 'owners',),
        'delete': ('managers', ),
        }

    ## attributes
    num              = String(required=True, fulltextindexed=True, indexed=True,
                              constraints=[SizeConstraint(16)],
                              description=_('release number'))
    description      = RichString(fulltextindexed=True, default_format='text/rest',
                                  description=_('targets for this version'))
    starting_date   = Date(description=_('estimated starting date'))
    prevision_date   = Date(description=_('estimated publication date'))
    publication_date = Date(description=_('actual publication date'))

    ## relations
    version_of = SubjectRelation('Project', cardinality='1*', composite='object')
    todo_by = SubjectRelation('CWUser',
                              constraints=[RQLConstraint('O in_state ST, ST name "activated"'),
                                           RQLVocabularyConstraint('O in_group G, G name "staff"')],
                              description=_('users responsible for this version'))
    conflicts = SubjectRelation('Version',
                                constraints=[RQLVocabularyConstraint('S version_of PS, O version_of PO, EXISTS(PS uses PO) OR EXISTS(PO subproject_of PS)')],
                                description=_('client project\'s version conflicting with this version'))
    depends_on = SubjectRelation('Version',
                                 constraints=[RQLVocabularyConstraint('S version_of PS, O version_of PO, EXISTS(PS uses PO) OR EXISTS(PO subproject_of PS)')],
                                 description=_('client project\'s version dependency for this version'))


class Ticket(WorkflowableEntityType):
    """a ticket is representing some kind of work to do (or done) on a project
    (bug fix, feature request...)
    """
    __permissions__ = {
        'read':   ('managers', 'users', 'guests',),
        # right to add Ticket is actually granted on the concerns relation
        'add':    ('managers', 'users',),
        # client can only modify tickets when they are in the "open" state
        'update': ('managers', 'staff', xexpr('developer'),
                   ERQLExpression(xperm('client')+', X in_state S, S name "created"'),),
        'delete': ('managers', ),
        }

    ## attributes
    title = String(required=True, fulltextindexed=True,
                   constraints=[SizeConstraint(128)])

    description = RichString(fulltextindexed=True, default_format='text/rest',
                             description=_("give any useful information (to "
                                           "describe a bug think to give your "
                                           "configuration and some way to reproduce"
                                           " it whenever possible)"))
    priority = String(required=True, internationalizable=True,
                      vocabulary=[_('important'), _('normal'), _('minor')],
                      description=_("importance"), default='normal')
    type = String(required=True, internationalizable=True,
                  vocabulary=[_('bug'), _('enhancement'), _('task')],
                  description=_("type"), default='bug')

    ## relations
    concerns = SubjectRelation('Project', cardinality='1*', composite='object')
    appeared_in = SubjectRelation(('Version'), cardinality='?*',
                                   constraints=[RQLConstraint('S concerns P, O version_of P'),
                                                RQLVocabularyConstraint('O in_state ST, ST name "published"')],
                                   description=_("version in which a bug has been encountered"))
    done_in = SubjectRelation(('Version'), cardinality='?*',
                              constraints=[RQLConstraint('S concerns P, O version_of P'),
                                           RQLVocabularyConstraint('O in_state ST, NOT ST name "published"')],
                              description=_("version in which this ticket will be / has been  done"))

    identical_to = SubjectRelation('Ticket',
                                   constraints=[RQLConstraint('S concerns P, O concerns P')])
    depends_on = SubjectRelation('Ticket',
                                 description=_("ticket which has to be done to complete this one"),
                                 constraints=[RQLVocabularyConstraint(
                                     'S concerns P, EXISTS(O concerns P) '
                                     'OR EXISTS(O concerns PS, '
                                     'EXISTS(P uses P2) OR '
                                     'EXISTS(P2 subproject_of P))')])





# extra relation definitions ##################################################

class see_also(RelationDefinition):
    symmetric = True
    subject = ('Project', 'Ticket')
    object = ('Project', 'Ticket')


class require_permission(RelationDefinition):
    subject = ('Project', 'Version', 'Ticket')
    object = 'CWPermission'

class has_group_permission(RelationDefinition):
    """short cut relation for 'U in_group G, P require_group G' for efficiency
    reason. This relation is set automatically, you should not set this.
    """
    subject = 'CWUser'
    object = 'CWPermission'
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers',),
        'delete': ('managers',),
        }

# relation types permissions ##################################################

class subproject_of(RelationType):
    """link a ticket to the version in which it has been done"""
    inlined = True
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers', 'staff', RRQLExpression('U has_update_permission O', 'O')),
        'delete': ('managers', 'staff', RRQLExpression('U has_update_permission O', 'O')),
    }

class version_of(RelationType):
    """link a version to its project. A version is necessarily linked to one and
    only one project.
    """
    inlined = True
    __permissions__ = {
        'read':   ('managers', 'users', 'guests',),
        'add':    ('managers', 'staff', oexpr('client'),),
        'delete': ('managers', ),
        }

class todo_by(RelationType):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers',),
        'delete': ('managers',),
        }

class conflicts(RelationType):
    symmetric = True
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers', 'staff', sexpr('developer'),),
        'delete': ('managers', 'staff', sexpr('developer'),),
        }

class concerns(RelationType):
    """link a ticket to its project. A ticket is necessarily linked to one and
    only one project.
    """
    inlined = True
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers', 'staff', oexpr('developer', 'client'),),
        'delete': ('managers', 'staff',),
    }

class done_in(RelationType):
    """link a ticket to the version in which it has been done"""
    inlined = True
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        # XXX should require permission on the ticket or on the version, or both?
        # XXX should check the state of a ticket
        'add':    ('managers', 'staff', restricted_oexpr('O in_state ST, ST name "planned"', 'client'),),
        'delete': ('managers', 'staff', restricted_oexpr('O in_state ST, ST name "planned"', 'client'),),
    }

class appeared_in(RelationType):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        # XXX should require permission on the ticket or on the version, or both?
        'add':    ('managers', 'staff', restricted_oexpr('O in_state ST, ST name "published"', 'developer', 'client'),),
        'delete': ('managers', 'staff', restricted_oexpr('O in_state ST, ST name "published"', 'developer', 'client'),),
    }

class depends_on(RelationType):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers', 'staff', sexpr('developer', 'client'),),
        'delete': ('managers', 'staff', sexpr('developer', 'client'),),
    }




"""some utilities for testing tracker security

:organization: Logilab
:copyright: 2008-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from os.path import dirname, join, abspath

from logilab.common.decorators import cached

from cubicweb import Unauthorized, ValidationError
from cubicweb.devtools import (BaseApptestConfiguration, TestServerConfiguration,
                               ApptestConfiguration, init_test_database)
from cubicweb.devtools.testlib import CubicWebTC

def ptransitions(entity):
    return sorted(tr.name for tr in entity.possible_transitions())

def create_project_rql(pname, description=None):
    return 'INSERT Project X: X name %(name)s, X description %(descr)s', \
           {'name': unicode(pname), 'descr': unicode(description)}

def create_version_rql(num, pname):
    return 'INSERT Version X: X num %(num)s, X version_of P '\
           'WHERE P name %(name)s', \
           {'num': unicode(num), 'name': unicode(pname)}

def create_ticket_rql(title, pname):
    return 'INSERT Ticket X: X title %(title)s, X concerns P '\
           'WHERE P name %(name)s', \
           {'title': unicode(title), 'name': unicode(pname)}


class TrackerTCMixIn(object):

    def create_project(self, pname, description=None):
        return self.execute(*create_project_rql(pname, description))

    def create_version(self, num, pname='cubicweb'):
        return self.execute(*create_version_rql(num, pname))

    def create_ticket(self, title, vnum=None, pname='cubicweb'):
        rset = self.execute(*create_ticket_rql(title, pname))
        if vnum:
            self.execute('SET X done_in V WHERE X eid %(x)s, V num %(num)s',
                         {'x': rset[0][0], 'num': vnum})
        return rset

    def grant_permission(self, project, group, pname, plabel=None, etype='Project'):
        """insert a permission on a project. Will have to commit the main
        connection to be considered
        """
        pname = unicode(pname)
        plabel = plabel and unicode(plabel) or unicode(group)
        peid = self.session.unsafe_execute(
            'INSERT CWPermission X: X name %%(pname)s, X label %%(plabel)s,'
            'X require_group G, '
            'P require_permission X '
            'WHERE G name %%(group)s, P is %s, P name %%(project)s' % etype,
            locals())[0][0]
        return peid

    def assertModificationDateGreater(self, entity, olddate):
        entity.pop('modification_date', None)
        self.failUnless(entity.modification_date > olddate)

    def assertPossibleTransitions(self, entity, expected):
        self.assertListEquals(ptransitions(entity), sorted(expected))


class TrackerBaseTC(TrackerTCMixIn, CubicWebTC):
    def setup_database(self):
        self.cubicweb = self.request().create_entity('Project', name=u'cubicweb',
                                        description=u"cubicweb c'est beau")


class SecurityTC(TrackerTCMixIn, CubicWebTC):
    repo_config = BaseApptestConfiguration('data')
    _initialized = False
    def setUp(self):
        CubicWebTC.setUp(self)
        self.__cnxs = {'admin': self.cnx}
        # trick to avoid costly initialization for each test
        if not SecurityTC._initialized:
            # implicitly test manager can add some entities
            self.__class__.cubicweb = self.execute(*create_project_rql("cubicweb")).get_entity(0, 0)
            self.execute(*create_project_rql("projet2"))
            self.execute('INSERT CWGroup X: X name "cubicwebdevelopers"')
            self.execute('INSERT CWGroup X: X name "cubicwebclients"')
            self.grant_permission('cubicweb', 'cubicwebdevelopers', u'developer')
            self.grant_permission('cubicweb', 'cubicwebclients', u'client')
            self.commit()
            self.create_user('stduser')
            self.create_user('staffuser', groups=('users', 'staff',))
            self.create_user('prj1developer', groups=('users', 'cubicwebdevelopers',))
            self.create_user('prj1client', groups=('users', 'cubicwebclients'))
            self.maxeid = self.execute('Any MAX(X)')[0][0]
            #SecurityTC._initialized = True
            cachedperms = self.execute('Any UL, PN WHERE U has_group_permission P, U login UL, P label PN')
            self.assertEquals(len(cachedperms), 2)
            self.assertEquals(dict(cachedperms),
                              {'prj1developer': 'cubicwebdevelopers', 'prj1client': 'cubicwebclients'})

    def mylogin(self, user):
        if not user in self.__cnxs:
            self.__cnxs[user] = self.login(user)
        return self.__cnxs[user]

    def _test_tr_fail(self, user, x, trname):
        cnx = self.mylogin(user)
        try:
            entity = cnx.request().entity_from_eid(x)
            # if the user can't see entity x, Unauthorized is raised, else if he
            # can't pass the transition, Validation is raised
            self.assertRaises((Unauthorized, ValidationError),
                              entity.fire_transition, trname)
        finally:
            cnx.rollback()

    def _test_tr_success(self, user, x, trname):
        cnx = self.mylogin(user)
        try:
            entity = cnx.request().entity_from_eid(x)
            entity.fire_transition(trname)
            cnx.commit()
        except:
            cnx.rollback()
            raise

#from unittest_querier_planner import do_monkey_patch, undo_monkey_patch

class ThreeSourcesConfiguration(ApptestConfiguration):
    @cached
    def sources(self):
        res = TestServerConfiguration.sources(self)
        res['extern'] = {'adapter':    'pyrorql',
                         'pyro-ns-id': 'extern',
                         'cubicweb-user': 'admin',
                         'cubicweb-password': 'gingkow',
                         'mapping-file': abspath(join(dirname(__file__), 'test', 'data', 'extern_mapping.py'))
                         #'base-url': '',
                         }
        res['extern2'] = {'adapter':    'pyrorql',
                         'pyro-ns-id': 'extern2',
                         'cubicweb-user': 'admin',
                         'cubicweb-password': 'gingkow',
                         'mapping-file': abspath(join(dirname(__file__), 'test', 'data', 'extern_mapping.py'))
                         #'base-url': '',
                         }
        if self._enabled_sources is None:
            return res
        return dict((uri, config) for uri, config in res.items()
                    if uri in self._enabled_sources)


class ExternalSourceConfiguration(TestServerConfiguration):
    @cached
    def sources(self):
        res = {'admin': {'login': 'admin', 'password': 'gingkow'}}
        res['system'] = {'adapter':     'native',
                         'db-driver':   'sqlite',
                         'db-name':     'tmpdb-extern',
                         'db-encoding': 'utf8',
                         'db-host':     '',
                         'db-user':     'admin',
                         'db-password': 'gingkow',
                         }
        return res

class External2SourceConfiguration(TestServerConfiguration):
    @cached
    def sources(self):
        res = {'admin': {'login': 'admin', 'password': 'gingkow'}}
        res['system'] = {'adapter':     'native',
                         'db-driver':   'sqlite',
                         'db-name':     'tmpdb-extern2',
                         'db-encoding': 'utf8',
                         'db-host':     '',
                         'db-user':     'admin',
                         'db-password': 'gingkow',
                         }
        return res

def ms_test_init():
    repo2, cnx2 = init_test_database(config=ExternalSourceConfiguration('data'))
    cu = cnx2.cursor()
    ep1eid = cu.execute('INSERT Project X: X name "Extern project 1"')[0][0]
    ep1v1eid = cu.execute('INSERT Version V: V num "0.1.0", V version_of X WHERE X name "Extern project 1"')[0][0]
    ep2eid = cu.execute('INSERT Project X: X name "Extern project 2"')[0][0]
    ep2v1eid = cu.execute('INSERT Version V: V num "0.1.0", V version_of X WHERE X name "Extern project 2"')[0][0]
    cnx2.commit()

    repo3, cnx3 = init_test_database(config=External2SourceConfiguration('data'))
    cu3 = cnx3.cursor()
    e2p1eid = cu3.execute('INSERT Project X: X name "Extern2 project 1"')[0][0]
    cnx3.commit()

    # XXX, access existing connection, no pyro connection
    from cubicweb.server.sources.pyrorql import PyroRQLSource
    PyroRQLSource.get_connection = lambda x: x.uri == 'extern' and cnx2 or cnx3
    # necessary since the repository is closing its initial connections pool though
    # we want to keep cnx2 valid
    from cubicweb.dbapi import Connection
    Connection.close = lambda x: None

    eids = {'ep1eid': ep1eid,
            'ep1v1eid': ep1v1eid,
            'ep2eid': ep2eid,
            'ep2v1eid': ep2v1eid,
            'e2p1eid': e2p1eid
            }
    return repo2, cnx2, repo3, cnx3, eids


class BaseMSTC(CubicWebTC):
    config = ThreeSourcesConfiguration('data')

    def setUp(self):
        CubicWebTC.setUp(self)
        self.extern = self.repo.sources_by_uri['extern']
        # trigger import from external source since sqlite doesn't support
        # concurrent access to the database
        self.execute('Any P, V WHERE V? version_of P')
        self.execute('State X')
        self.maxeid = self.execute('Any MAX(X)')[0][0]
        self.ip1eid = self.execute('INSERT Project X: X name "Intern project 1"')[0][0]
        self.ip1v1eid = self.execute('INSERT Version V: V num "0.1.0", V version_of X WHERE X name "Intern project 1"')[0][0]
        self.ip1t1eid = self.execute('INSERT Ticket X: X title "ticket project 1", X concerns P, X done_in V WHERE P eid %(p)s, V eid %(v)s',
                                     {'p': self.ip1eid, 'v': self.ip1v1eid})[0][0]
        self.ip2eid = self.execute('INSERT Project X: X name "Intern project 2"')[0][0]
        self.ip2v1eid = self.execute('INSERT Version V: V num "0.1.0", V version_of X WHERE X name "Intern project 2"')[0][0]
        self.ip2t1eid = self.execute('INSERT Ticket X: X title "ticket project 2", X concerns P, X done_in V, X depends_on T WHERE P eid %(p)s, V eid %(v)s, T eid %(t)s',
                                     {'p': self.ip2eid, 'v': self.ip2v1eid, 't': self.ip1t1eid})[0][0]
        self.ip3eid = self.execute('INSERT Project X: X name "Intern project 3"')[0][0]
        self.commit()

    def extid2eid(self, extid, etype):
        return self.repo.extid2eid(self.extern, str(extid), etype, self.session)

    def fire_transition(self, eid, trname):
        e = self.execute('Any X WHERE X eid %(x)s', {'x': eid}, 'x').get_entity(0, 0)
        e.fire_transition(trname)

    def change_state(self, eid, stname):
        e = self.execute('Any X WHERE X eid %(x)s', {'x': eid}, 'x').get_entity(0, 0)
        e.change_state(stname)




"""views for Project entities

:organization: Logilab
:copyright: 2006-2009 LOGILAB S.A. (Paris, FRANCE), license is LGPL.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from logilab.mtconverter import xml_escape
from logilab.common import table as _table

from cubicweb.schema import display_name
from cubicweb.view import EntityView, EntityStartupView
from cubicweb.selectors import (implements, has_related_entities, none_rset,
                                anonymous_user, authenticated_user, score_entity)
from cubicweb import tags
from cubicweb.web import component, uicfg
from cubicweb.web.views import tabs, baseviews

_pvdc = uicfg.primaryview_display_ctrl
_pvdc.tag_attribute(('Project', 'summary'), {'showlabel': False})
_pvdc.tag_attribute(('Project', 'description'), {'showlabel': False})

# primary view and tabs ########################################################

class ProjectPrimaryView(tabs.TabbedPrimaryView):
    __select__ = implements('Project')

    tabs = [_('projectinfo_tab'), _('projectroadmap_tab'), _('projecttickets_tab')]
    default_tab = 'projectinfo_tab'


# configure projectinfotab
_pvs = uicfg.primaryview_section
_pvs.tag_attribute(('Project', 'name'), 'hidden')
# XXX keep '*', not Project to match other target types added in eg forge
_pvs.tag_subject_of(('Project', 'uses', '*'), 'attributes')
_pvs.tag_object_of(('Project', 'uses', '*'), 'attributes')
_pvs.tag_object_of(('Project', 'subproject_of', '*'), 'hidden')
_pvs.tag_subject_of(('*', 'subproject_of', 'Project'), 'hidden')

class ProjectInfoTab(tabs.PrimaryTab):
    __regid__ = 'projectinfo_tab'
    __select__ = implements('Project')

    title = None # should not appear in possible views


class ProjectRoadmapTab(EntityView):
    """display the latest published version and in preparation version"""
    __regid__ = 'projectroadmap_tab'
    __select__ = (anonymous_user() & implements('Project') &
                  has_related_entities('version_of', 'object'))

    title = None # should not appear in possible views

    def cell_call(self, row, col):
        self.cw_rset.get_entity(row, col).view('roadmap', w=self.w)


class ProjectTicketsTab(EntityView):
    __regid__ = 'projecttickets_tab'
    __select__ = implements('Project')

    title = None # should not appear in possible views

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        # optimization: prefetch project's to fill entity / relations cache
        entity.reverse_version_of
        rset = self._cw.execute(entity.tickets_rql(limit=1))
        self._cw.form['actualrql'] = entity.active_tickets_rql()
        self.wview('editable-initialtable', rset, 'null',
                   subvid='incontext', displayactions=1,
                   divid='tickets%s' % entity.eid,
                   displaycols=range(entity.tickets_rql_nb_displayed_cols))


# contextual components ########################################################

class ProjectRoadmapVComponent(component.EntityVComponent):
    """display the latest published version and in preparation version"""
    __regid__ = 'roadmap'
    __select__ = (component.EntityVComponent.__select__ &
                  authenticated_user() & implements('Project') &
                  has_related_entities('version_of', 'object'))
    context = 'navcontenttop'
    title = _('Version_plural')
    order = 10

    def cell_call(self, row, col, view=None):
        self.cw_rset.get_entity(row, col).view('roadmap', w=self.w)


class ProjectTreeVComponent(component.EntityVComponent):
    __regid__ = 'projecttree'
    title = _('project tree')
    __select__ = (component.EntityVComponent.__select__ &
                  implements('Project') &
                  score_entity(lambda x: x.children()))
    context = 'navcontentbottom'

    def cell_call(self, row, col, view=None):
        entity = self.cw_rset.get_entity(row, col)
        rset = entity.children(entities=False)
        if rset:
            self.w(u'<div class="section">')
            self.w(tags.h2(self._cw._('Project tree')))
            treeid = 'project_tree_%s' % entity.eid
            self.w(self.view('treeview', rset=rset, treeid=treeid,
                             initial_thru_ajax=True))
            self.w(u'</div>')


# secondary views ##############################################################

class ProjectRoadmapView(EntityView):
    """display the latest published version and in preparation version"""
    __regid__ = 'roadmap'
    __select__ = (implements('Project') &
                  has_related_entities('version_of', 'object'))
    title = None # should not appear in possible views
    rql = ('Any V,DATE ORDERBY version_sort_value(N) '
           'WHERE V num N, V prevision_date DATE, V version_of X, '
           'V in_state S, S name IN ("planned", "dev", "ready"), '
           'X eid %(x)s')

    def cell_call(self, row, col):
        self.w(u'<div class="section">')
        entity = self.cw_rset.get_entity(row, col)
        currentversion = entity.latest_version()
        if currentversion:
            self.w(self._cw._('latest published version:'))
            self.w(u'&nbsp;')
            currentversion.view('incontext', w=self.w)
            self.w(u'<br/>')
        rset = self._cw.execute(self.rql, {'x': entity.eid}, 'x')
        if rset:
            self.wview('ic_progress_table_view', rset)
        allversionsrql = entity.related_rql('version_of', 'object') % {'x': entity.eid}
        self.w('<a href="%s">%s</a>'
               % (xml_escape(self._cw.build_url(vid='list', rql=allversionsrql)),
                  self._cw._('view all versions')))
        self.w(u'</div>')


class ProjectOutOfContextView(baseviews.OutOfContextView):
    """project's out of context view display project's url, which is not
    displayed in the one line / text views
    """
    __select__ = implements('Project')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(tags.a(entity.name, href=entity.absolute_url()))
        # no summary on ext project
        if getattr(entity, 'summary', None):
            self.w(u'&nbsp;')
            self.w(xml_escape(entity.summary))


# other views ##################################################################

class ProjectStatsView(EntityStartupView):
    """Some statistics : how many bugs, sorted by status, indexed by projects
    """
    __regid__ = 'stats'
    __select__ = none_rset() | implements('Project')
    title = _('projects statistics')
    default_rql = 'Any P,PN WHERE P name PN, P is Project'

    def call(self, sort_col=None):
        w = self.w
        req = self._cw
        req.add_css('cubes.tracker.stats.css')
        if self.cw_rset is None:
            self.cw_rset = req.execute(self.default_rql)
        table = _table.Table()
        statuslist = [row[0] for row in self._cw.execute('DISTINCT Any N WHERE S name N, X in_state S, X is Ticket')]
        severities = ['minor', 'normal', 'important']
        table.create_columns(statuslist + severities + ['Total'])
        nb_cols = len(table.col_names)
        # create a stylesheet to compute sums over rows and cols
        stylesheet = _table.TableStyleSheet()
        # fill table
        i = -1
        for row in self.cw_rset:
            i += 1
            eid = row[0]
            row = []
            total = 0
            for status in statuslist:
                rql = "Any COUNT(A) WHERE A is Ticket, A concerns P, A in_state S, S name %(s)s, P eid %(x)s"
                nbtickets = req.execute(rql, {'x': eid, 's': status}, 'x', build_descr=0)[0][0]
                total += nbtickets
                row.append(nbtickets)
            for severity in severities:
                rql = "Any COUNT(A) WHERE A is Ticket, A concerns P, A priority %(p)s, P eid %(x)s"
                nbtickets = req.execute(rql, {'x': eid, 'p': severity}, 'x', build_descr=0)[0][0]
                total += nbtickets
                row.append(nbtickets)
            row.append(total)
            table.append_row(row, xml_escape(self.cw_rset.get_entity(i, 0).name))
            assert len(row) == nb_cols
        # sort table according to sort_col if wanted
        sort_col = sort_col or self._cw.form.get('sort_col', '')
        if sort_col:
            table.sort_by_column_id(sort_col, method='desc')
        else:
            table.sort_by_column_index(0)
        # append a row to compute sums over rows and add appropriate
        # stylesheet rules for that
        if len(self.cw_rset) > 1:
            table.append_row([0] * nb_cols, 'Total')
            nb_rows = len(table.row_names)
            for i in range(nb_cols):
                stylesheet.add_colsum_rule((nb_rows-1, i), i, 0, nb_rows-2)
            table.apply_stylesheet(stylesheet)
        # render the table
        w(u'<table class="stats" cellpadding="5">')
        w(u'<tr>')
        for col in [''] + table.col_names:
            url = self._cw.build_url(vid='stats', sort_col=col,
                                 __force_display=1,
                                 rql=self.cw_rset.printable_rql())
            self.w(u'<th><a href="%s">%s</a></th>\n' % (xml_escape(url), col))
        self.w(u'</tr>')
        for row_name, row, index in zip(table.row_names, table.data,
                                        xrange(len(table.data))):
            if index % 2 == 0:
                w(u'<tr class="alt0">')
            else:
                w(u'<tr class="alt1">')
            if index == len(table.data) - 1:
                w(u'<td>%s</td>' % row_name)
            else:
                url = self._cw.build_url('project/%s' % self._cw.url_quote(row_name))
                self.w(u'<td><a href="%s">%s</a></td>' % (xml_escape(url), row_name))
            for cell_data in row:
                w(u'<td>%s</td>' % cell_data)
            w(u'</tr>')
        w(u'</table>')


class SubscribeToReleasesComponent(component.EntityVComponent):
    """link to subscribe to rss feed for published versions of project
    """
    __regid__ = 'projectreleasesubscriberss'
    __select__ = (component.EntityVComponent.__select__ &
                  implements('Project'))
    context = 'ctxtoolbar'

    def cell_call(self, row, col, view=None):
        entity = self.cw_rset.get_entity(row, col)
        label = self._cw._(u'Subscribe to project releases')
        logo = u'<img src="%s" alt="%s"/>' % (
            self._cw.external_resource('RSS_LOGO_16'), label)
        rql = 'Any V, VN ORDERBY VN DESC WHERE V version_of P, P eid %s, ' \
              'V in_state S, S name "published", V num VN' % entity.eid
        # XXX <project>/versions/rss ?
        url = self._cw.build_url('view', vid='rss', rql=rql)
        self.w(u'<a href="%s" title="%s" class="toolbarButton">%s</a>' % (
            xml_escape(url), label, logo))





"""xml/rdf views for tracker related entities (doap, linked data)

:organization: Logilab
:copyright: 2001-2009 LOGILAB S.A. (Paris, FRANCE), license is LGPL.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from logilab.mtconverter import xml_escape

from cubicweb.view import EntityView
from cubicweb.selectors import implements

class LinkedDataView(EntityView):
    __regid__ = 'linked_data'
    __select__ = implements('Project')

    title = _('linked data')
    templatable = False
    content_type = 'text/xml'
    item_vid = 'linked_data_item'

    def call(self):
        self.w(u'<?xml version="1.0" encoding="%s"?>\n' % self._cw.encoding)
        self.w(u'''<rdf:RDF
             xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
             xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#"
             xmlns:owl="http://www.w3.org/2002/07/owl#"
             xmlns:foaf="http://xmlns.com/foaf/0.1/"
             xmlns:doap="http://usefulinc.com/ns/doap#">\n''')
        for i in xrange(self.cw_rset.rowcount):
            self.cell_call(row=i, col=0)
        self.w(u'</rdf:RDF>\n')

    def cell_call(self, row, col):
        self.wview(self.item_vid, self.cw_rset, row=row, col=col)


class LinkedDataProjectItemView(EntityView):
    __regid__ = 'linked_data_item'
    __select__ = implements('Project')

    def cell_call(self, row, col):
        '''display all project attribut and project dependencies and external project (in doap format) if
        it is related to'''
        entity = self.cw_rset.complete_entity(row, col)
        self.w(u' <doap:Project rdf:about="%s">\n' % xml_escape(entity.absolute_url()))
        self.common_doap(entity)
        self.w(u' </doap:Project>\n')

    def project_references(self, entity, refattr):
        refs = getattr(entity, refattr)
        if refs:
            self.w(u'<%s>' % refattr)
            for ref in refs:
                self.w(u'<doap:Project rdf:resource="%s"/>'% xml_escape(ref.absolute_url()))
            self.w(u'</%s>' % refattr)

    def common_doap(self, entity):
        self.w(u'  <doap:name>%s</doap:name>\n'% xml_escape(entity.dc_title()))
        self.w(u'  <doap:created>%s</doap:created>\n' % (entity.creation_date.strftime('%Y-%m-%d')))
        if entity.summary:
            self.w(u'  <doap:shortdesc>%s</doap:shortdesc>\n' % xml_escape(entity.summary))
        descr = entity.printable_value('description', format='text/plain')
        if descr:
            self.w(u'  <doap:description>%s</doap:description>\n' % xml_escape(descr))



class ProjectDoapItemView(LinkedDataProjectItemView):
    __regid__ = 'doapitem'
    __select__ = implements('Project')

    def cell_call(self, row, col):
        """ element as an item for an doap description """
        entity = self.cw_rset.complete_entity(row, col)
        self.w(u'<doap:Project rdf:about="%s">\n' % xml_escape(entity.absolute_url()))
        self.common_doap(entity)
        # version
        ver = entity.latest_version()
        if ver:
            self.w(u'  <doap:release>\n')
            ver.view('doapitem', w=self.w)
            self.w(u'  </doap:release>\n')
        # repository
        self.w(u'</doap:Project>\n')


class VersionDoapItemView(EntityView):
    __regid__ = 'doapitem'
    __select__ = implements('Version')

    def cell_call(self, row, col):
        """ element as an item for an doap description """
        entity = self.cw_rset.complete_entity(row, col)
        self.w(u'<doap:Version>\n')
        self.w(u'  <doap:revision>%s</doap:revision>\n' % xml_escape(entity.num) )
        self.w(u'  <doap:created>%s</doap:created>\n' % entity.dc_date('%Y-%m-%d'))
        self.w(u'</doap:Version>\n')


class DoapView(LinkedDataView):
    __regid__ = 'doap'
    title = _('doap')
    item_vid = 'doapitem'
    __select__ = implements('Project', 'Version')






"""tracker extproject and project entity classes

:organization: Logilab
:copyright: 2006-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from warnings import warn

from logilab.common.decorators import cached

from cubicweb.interfaces import ITree
from cubicweb.mixins import TreeMixIn
from cubicweb.entities import AnyEntity, fetch_config

from cubes.tracker.entities import fixed_orderby_rql

class Project(TreeMixIn, AnyEntity):
    __regid__ = 'Project'
    __implements__ = AnyEntity.__implements__ + (ITree,)
    __permissions__ = ('developer', 'client')
    fetch_attrs, fetch_order = fetch_config(('name', 'description', 'description_format',
                                             'summary'))

    TICKET_DEFAULT_STATE_RESTR = 'S name IN ("created","identified","released","scheduled")'

    tree_attribute = 'subproject_of'
    parent_target = 'subject'
    children_target = 'object'

    def dc_title(self):
        return self.name

    def dc_long_title(self):
        if self.summary:
            return u'%s (%s)' % (self.name, self.summary)
        return self.name

    @property
    def project(self):
        """project item interface"""
        return self

    def latest_version(self, states=('published',), reverse=None):
        """returns the latest version(s) for the project in one of the given
        states.

        when no states specified, returns the latest published version.
        """
        order = 'DESC'
        if reverse is not None:
            warn('reverse argument is deprecated',
                 DeprecationWarning, stacklevel=1)
            if reverse:
                order = 'ASC'
        rset = self.versions_in_state(states, order, True)
        if rset:
            return rset.get_entity(0, 0)
        return None

    @cached
    def versions_in_state(self, states, order='ASC', limit=False):
        """returns version(s) for the project in one of the given states, sorted
        by version number.

        If limit is true, limit result to one version.
        If reverse, versions are returned from the smallest to the greatest.

        NOTE: /!\ DO NOT CALL this method with named parameter: `cached`
              implementation doesn't support it yet
        """
        if limit:
            order += ' LIMIT 1'
        rql = 'Any V,N ORDERBY version_sort_value(N) %s ' \
              'WHERE V num N, V in_state S, S name IN (%s), ' \
              'V version_of P, P eid %%(p)s' % (order, ','.join(repr(s) for s in states))
        return self._cw.execute(rql, {'p': self.eid})

    # number of columns to display
    tickets_rql_nb_displayed_cols = 7
    sort_defs = (('in_state', 'S'), ('num', 'VN'), ('priority', 'PR'))
    def tickets_rql(self, limit=None):
        return ('Any B, NOW - CD, NOW - BMD, U,PR,S,V,BD,VN,P,BT,CD,BMD,UL '
                'GROUPBY B,CD,PR,S,V,U,VN,BD,P,BT,BMD,UL '
                '%s %s '
                'WHERE B priority PR, B in_state S, B creation_date CD, '
                'B description BD, '
                'B title BT, B modification_date BMD, '
                'B done_in V?, V num VN, '
                'B created_by U?, U login UL, B concerns P, P eid %s'
               ) % (fixed_orderby_rql(self.sort_defs),
                    limit and ' LIMIT %r' % limit or '',
                    self.eid)
        return rql

    def active_tickets_rql(self):
        return '%s, %s' % (self.tickets_rql(), self.TICKET_DEFAULT_STATE_RESTR)



"""Custom form for tracker

:organization: Logilab
:copyright: 2003-2010 LOGILAB S.A. (Paris, FRANCE), license is LGPL.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from datetime import datetime

from cubicweb.selectors import implements, match_transition
from cubicweb.web import uicfg, formfields
from cubicweb.web.views import workflow, editviews


_afs = uicfg.autoform_section
_afs.tag_attribute(('Version', 'publication_date'), 'main', 'hidden')
_afs.tag_attribute(('Version', 'starting_date'), 'main', 'attributes')
_afs.tag_attribute(('Version', 'prevision_date'), 'main', 'attributes')
_afs.tag_subject_of(('Version', 'version_of', 'Project'), 'main', 'hidden')
_afs.tag_object_of(('Version', 'version_of', 'Project'), 'main', 'hidden')
_afs.tag_subject_of(('Ticket', 'concerns', 'Project'), 'main', 'attributes')
_afs.tag_object_of(('Ticket', 'concerns', 'Project'), 'main', 'hidden')
_afs.tag_subject_of(('Ticket', 'done_in', 'Version'), 'main', 'attributes')

def ticket_done_in_choices(form, field):
    entity = form.edited_entity
    # first see if its specified by __linkto form parameters
    linkedto = formfields.relvoc_linkedto(entity, 'done_in', 'subject')
    if linkedto:
        return linkedto
    # it isn't, check if the entity provides a method to get correct values
    vocab = formfields.relvoc_init(entity, 'done_in', 'subject')
    if not entity.has_eid():
        peids = entity.linked_to('concerns', 'subject')
        peid = peids and peids[0]
    else:
        peid = entity.project.eid
    if peid:
        rschema = form._cw.vreg.schema['done_in'].rdef('Ticket', 'Version')
        rset = form._cw.execute(
            'Any V, VN, P ORDERBY version_sort_value(VN) '
            'WHERE V version_of P, P eid %(p)s, V num VN, '
            'V in_state ST, NOT ST name "published"', {'p': peid}, 'p')
        vocab += [(v.view('combobox'), v.eid) for v in rset.entities()
                  if rschema.has_perm(form._cw, 'add', toeid=v.eid)]
    return vocab

_affk = uicfg.autoform_field_kwargs
_affk.tag_attribute(('Ticket', 'priority'), {'sort': False})
_affk.tag_subject_of(('Ticket', 'done_in', '*'), {'sort': False,
                                                  'choices': ticket_done_in_choices})


class VersionChangeStateForm(workflow.ChangeStateForm):
    __select__ = implements('Version') & match_transition('publish')

    publication_date = formfields.DateField(eidparam=True, role='subject',
                                            fallback_on_none_attribute=True,
                                            value=lambda form: datetime.now())


class ShortComboboxView(editviews.ComboboxView):
    """by default combobox view is redirecting to textoutofcontext view
    but in the case of projects we want a shorter view
    """
    __select__ = implements('Project')
    def cell_call(self, row, col):
        self.w(self.cw_rset.get_entity(row, col).dc_title())



from datetime import timedelta

from cubicweb import MultiSourcesError
from cubicweb.devtools.testlib import CubicWebTC

from cubes.tracker.testutils import BaseMSTC, ms_test_init

repo2, cnx2, repo3, cnx3, eids = ms_test_init()
globals().update(eids)

class ThreeSourcesTC(BaseMSTC):

    def test_projects_list(self):
        rset = self.execute('Project P')
        self.assertEquals(len(rset), 6)

    def test_nonregr1(self):
        rset = self.execute('Any X WHERE X eid %(x)s, V version_of X, V in_state S, S name "published"',
                            {'x': self.ip1eid}, 'x')
        self.assertEquals(rset.rows, [])
        self.change_state(self.ip1v1eid, 'published')
        rset = self.execute('Any X WHERE X eid %(x)s, V version_of X, V in_state S, S name "published"',
                            {'x': self.ip1eid}, 'x')
        self.assertEquals(len(rset.rows), 1)
        self.assertEquals(rset.rows[0][0], self.ip1eid)
        rset = self.execute('Any X WHERE X eid %(x)s, V version_of X, V in_state S, S name "published"',
                            {'x': self.ip2eid}, 'x')
        self.assertEquals(rset.rows, [])
        self.change_state(self.ip2v1eid, 'published')
        rset = self.execute('Any X WHERE X eid %(x)s, V version_of X, V in_state S, S name "published"',
                            {'x': self.ip2eid}, 'x')
        self.assertEquals(len(rset.rows), 1)
        # same test with entities from the external source
        iep1eid = self.extid2eid(ep1eid, 'Project')
        iep1v1eid = self.extid2eid(ep1v1eid, 'Project')
        rset = self.execute('Any X WHERE X eid %(x)s, V version_of X, V in_state S, S name "published"',
                            {'x': iep1eid}, 'x')
        self.assertEquals(rset.rows, [])
        # XXX don't work anymore since changing state is done by adding TrInfo which isn't currently
        #     imported from external sources
        self.assertRaises(MultiSourcesError, self.change_state, iep1v1eid, 'published')
        #self.change_state(iep1v1eid, 'published')
        # rset = self.execute('Any X WHERE X eid %(x)s, V version_of X, V in_state S, S name "published"',
        #                     {'x': iep1eid}, 'x')
        # self.assertEquals(len(rset.rows), 1, rset)
        # self.assertEquals(rset.rows[0][0], iep1eid)
        # iep2eid = self.extid2eid(ep2eid, 'Project')
        # iep2v1eid = self.extid2eid(ep2v1eid, 'Project')
        # rset = self.execute('Any X WHERE X eid %(x)s, V version_of X, V in_state S, S name "published"',
        #                     {'x': iep2eid}, 'x')
        # self.assertEquals(rset.rows, [])
        # self.change_state(iep2v1eid, 'published')
        # rset = self.execute('Any X WHERE X eid %(x)s, V version_of X, V in_state S, S name "published"',
        #                     {'x': iep2eid}, 'x')
        # self.assertEquals(len(rset.rows), 1)

    def test_nonregr2(self):
        ip2v1 = self.execute('Any X WHERE X eid %(x)s', {'x': self.ip2v1eid}, 'x').get_entity(0, 0)
        rset = ip2v1.depends_on_rset()
        self.assertEquals(len(rset.rows), 1)
        self.assertEquals(rset.rows[0][0], self.ip1v1eid)

    def test_nonregr3(self):
        self.execute('Any V,DATE ORDERBY version_sort_value(N) '
                     'WHERE V num N, V prevision_date DATE, V version_of X, '
                     'V in_state S, S name IN ("planned", "dev", "ready"), X eid %(x)s',
                     {'x': self.ip1eid})

    def test_nonregr4(self):
        # test we can a change a ticket state (and implicitly its version'state by commiting
        self.fire_transition(self.ip1t1eid, 'start')
        self.commit()
        ip1v1 = self.execute('Any X WHERE X eid %(x)s', {'x': self.ip1v1eid}, 'x').get_entity(0, 0)
        self.assertEquals(ip1v1.state, 'dev')

    def test_nonregr5(self):
        self.fire_transition(self.ip1t1eid, 'start')
        self.commit()
        rset = self.execute('Any X,U WHERE X in_state S, S name "in-progress", TR wf_info_for X, TR to_state S, TR owned_by U')
        self.failUnlessEqual(len(rset), 1, rset.rows)
        self.failUnlessEqual(rset[0], [self.ip1t1eid, self.session.user.eid])

    def test_nonregr7(self):
        rset = self.execute('Any B,TT,NOW - CD,PR,S,V '
                            'GROUPBY B,TT,CD,PR,S,V,VN '
                            'ORDERBY S, version_sort_value(VN), TT, priority_sort_value(PR) '
                            'LIMIT 1 '
                            'WHERE B type TT, B priority PR, B in_state S, B creation_date CD, B done_in V?, V num VN, B concerns P, P eid %s'
                            % self.ip1eid)
        rset.rows[0][4] = 'STATE'
        self.assertEquals(rset.rows, [[self.ip1t1eid, u'bug', timedelta(0), u'normal', 'STATE', self.ip1v1eid]])

    def test_nonregr8(self):
        self.execute('Any SN WHERE X has_text "toto", X in_state S, S name SN, X is IN (Project, Version, Transition)')

    def test_nonregr9(self):
        self.execute('DISTINCT Any AD,AE ORDERBY AE WHERE NOT S depends_on O, S eid %(x)s, O is Version, O in_state AD, AD name AE, AD modification_date AF',
                     {'x': self.ip1v1eid}, ('x',))

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



"""Restructured text view to export content of a tracker instance

:organization: Logilab
:copyright: 2007-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from logilab.common.textutils import normalize_text

from cubicweb.view import EntityView
from cubicweb.selectors import (implements, match_search_state, one_line_rset,
                                multi_lines_rset)
from cubicweb.entity import _marker
from cubicweb.web import action

class ReSTItemView(EntityView):
    __abstract__ = True
    __regid__ = 'docitem'

    title_underline = None # make pylint happy
    templatable = False

    def _title(self, title):
        self.w(u'%s\n%s\n' % (title, self.title_underline*len(title)))

    def paragraph_field(self, entity, attr, **kwargs):
        value = entity.printable_value(attr, format='text/plain', **kwargs)
        if value:
            attrformat = getattr(entity, '%s_format', 'text/plain')
            self.w(u'\n%s\n' % normalize_text(value, 80,
                                              rest=attrformat=='text/rest'))

    def field(self, entity, attr, value=_marker, **kwargs):
        if value is _marker:
            if attr == 'state':
                value = entity.printable_state
            else:
                value = entity.printable_value(attr, value, format='text/plain',
                                               **kwargs)
            attr = self._cw._(attr)
        if value:
            self.w(u':%s: %s\n' % (attr, value))

    def cell_call(self, row, col):
        """ReST view for Project entities"""
        entity = self.cw_rset.complete_entity(row, col)
        self._title(u'%s %s' % (entity.dc_type(), entity.dc_title()))
        self.render_attributes(entity)
        self.w(u'\n')
        self.render_child(entity)

    # default implementations, override when needed

    def render_title(self, entity):
        self._title(u'%s %s' % (entity.dc_type(), entity.dc_title()))

    fields = ()
    text_fields = ()
    def render_attributes(self, entity):
        for field in self.fields:
            self.field(entity, field, displaytime=False)
        for field in self.text_fields:
            self.paragraph_field(entity, field)

    def render_child(self, entity):
        pass


class ProjectDocumentItemView(ReSTItemView):
    """ReST view for Project entities"""
    __select__ = implements('Project')

    title_underline = '='
    fields = ['creation_date']
    text_fields = ['summary', 'description']

    def render_child(self, entity):
        # version (sort by ascending version.num)
        for ver in reversed(entity.reverse_version_of):
            ver.view('docitem', w=self.w)


class VersionDocumentItemView(ReSTItemView):
    """ReST view for Version entities"""
    __select__ = implements('Version')

    title_underline = '-'

    def render_title(self, entity):
        self._title(u'%s %s (%s)' % (entity.dc_type(), entity.num,
                                     entity.printable_state))

    def render_attributes(self, entity):
        self.paragraph_field(entity, 'description')
        if entity.state == 'published':
            self.field(entity, 'publication_date')
        else:
            self.field(entity, 'prevision_date')

    def render_child(self, entity):
        for ticket in self._cw.execute(entity.tickets_rql()).entities():
            ticket.view('docitem', w=self.w)


class TicketDocumentItemView(ReSTItemView):
    """ReST view for Ticket entities"""
    __select__ = implements('Ticket')

    title_underline = '~'
    fields = ['type', 'state']
    text_fields = ['description']

    def render_title(self, entity):
        self._title(u'%s #%s: %s [%s]' % (
            entity.dc_type(), entity.eid, entity.title,
            self._cw._(entity.priority)))


class DocumentView(EntityView):
    __regid__ = 'document'
    __select__ = implements('Project', 'Version', 'Ticket')

    title = _('document')
    templatable = False
    content_type = 'text/plain'

    def set_request_content_type(self):
        """overriden to set a .txt filename"""
        self._cw.set_content_type(self.content_type,
                                  filename='cubicwebexport.txt')

    def cell_call(self, row, col):
        self.wview('docitem', self.cw_rset, row=row, col=col)

    def call(self):
        self.w(u'.. -*- coding: %s -*-\n\n' % self._cw.encoding.lower())
        for i in xrange(self.cw_rset.rowcount):
            self.cell_call(row=i, col=0)


# actions ######################################################################

class ProjectVersionExportAction(action.Action):
    __regid__ = 'pvrestexport'
    __select__ = (match_search_state('normal') & one_line_rset() &
                  implements('Project', 'Version'))

    title = _('ReST export')
    order = 410

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid='document')


class TicketRESTExportAction(action.Action):
    __regid__ = 'ticketrestexport'
    __select__ = (match_search_state('normal') & multi_lines_rset() &
                  implements('Ticket'))
    title = _('ReST export')

    def url(self):
        return self._cw.build_url('view', rql=self.cw_rset.printable_rql(),
                                  vid='document')



"""tracker version entity class

:organization: Logilab
:copyright: 2006-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from datetime import date

from logilab.common.decorators import cached
from logilab.common.date import todate

from cubicweb.interfaces import (IMileStone, IPrevNext, ITimetableViews,
                                 ICalendarViews, ICalendarable)
from cubicweb.mixins import ProgressMixIn
from cubicweb.entities import AnyEntity

from cubes.tracker.entities import fixed_orderby_rql


class Version(ProgressMixIn, AnyEntity):
    __regid__ = 'Version'
    __implements__ = AnyEntity.__implements__ + (ICalendarViews, IMileStone, IPrevNext,
                                                 ITimetableViews, ICalendarable)
    fetch_attrs = ('num', 'description', 'in_state')

    FINISHED_STATES = (u'ready', u'published')
    PROGRESSING_STATES =  (u'dev',)
    PUBLISHED_STATES = (u'published')

    @classmethod
    def fetch_order(cls, attr, var):
        if attr == 'num':
            var = 'version_sort_value(%s)' % var
            return '%s DESC' % var
        return None
    fetch_unrelated_order = fetch_order

    def rest_path(self, use_ext_eid=False):
        return u'%s/%s' % (self.project.rest_path(), self.num)

    @property
    def project(self):
        """ProjectItem interface"""
        return self.version_of[0]

    # dublin core ##############################################################

    def dc_title(self, format='text/plain'):
        return self.num

    def dc_long_title(self):
        return u'%s %s' % (self.project.name, self.num)

    def dc_date(self, date_format=None):
        if self.publication_date:
            date = self.publication_date
        else:
            date = self.modification_date
        return self._cw.format_date(date, date_format=date_format)

    # version'specific logic ###################################################

    def depends_on_rset(self):
        """return a result set of versions on which this one depends or None"""
        rql = 'DISTINCT Version V WHERE MB done_in MV, MV eid %(x)s, '\
              'MB depends_on B, B done_in V, V version_of P, NOT P eid %(p)s'
        args = {'x': self.eid, 'p': self.project.eid}
        eids = set(str(r[0]) for r in self._cw.execute(rql, args))
        for row in self.related('depends_on'):
            eids.add(str(row[0]))
        if not eids:
            return None
        return self._cw.execute('Version V WHERE V eid in (%s)' % ','.join(eids))

    def next_version(self, states=('planned', 'dev')):
        """return the first version following this one which is in one of the
        given states
        """
        found = False
        for version in reversed(self.project.reverse_version_of):
            if found and (states is None or version.state in states):
                return version
            if version is self:
                found = True

    def open_tickets(self):
        return (ticket for ticket in self.reverse_done_in if ticket.is_open())

    # iprogress interface ############################################################
    def contractors(self):
        return [euser for euser in self.todo_by]

    def progress_info(self):
        """returns a dictionary describing load and progress of the version"""
        return {'notestimated': 0,
                'estimated': 0,
                'done': 0,
                'todo': 0 }

    # number of columns to display
    tickets_rql_nb_displayed_cols = 8
    defects_rql_nb_displayed_cols = 5
    sort_defs = (('in_state', 'S'), ('priority', 'PR'))

    def tickets_rql(self):
        return ('Any B,PR,S,U,TI,D,V '
                'GROUPBY B,PR,S,U,TI,D,V %s '
                'WHERE B priority PR, '
                'B in_state S, B created_by U?,'
                'B done_in V, V eid %s, '
                'B title TI, B description D'
                % (fixed_orderby_rql(self.sort_defs), self.eid))

    def defects_rql(self):
        return ('Any B,PR,S,U,TI,D,V '
                'GROUPBY B,PR,S,U,TI,D,V %s '
                'WHERE B priority PR, '
                'B in_state S, B created_by U?,'
                'B appeared_in V, V eid %s, '
                'B title TI, B description D'
                % (fixed_orderby_rql(self.sort_defs), self.eid))

    def sortvalue(self, rtype=None):
        if rtype is None or rtype == 'num':
            # small hack to add the project name in the sorted values so that
            # out-of-context views get sorted according to project names and
            # in-context views according to version number
            if rtype is None:
                value = [self.project.name]
            else:
                value = []
            for part in self.num.split('.'):
                try:
                    value.append(int(part))
                except ValueError:
                    value.append(part)
            return value
        return super(Version, self).sortvalue(rtype)

    @cached
    def start_date(self):
        """return start date of version, when first transition is done (to
        'dev' state)
        """
        # first tr_info is necessarily the transition to dev
        try:
            tr_info = self.reverse_wf_info_for[1]
            return todate(tr_info.creation_date)
        except IndexError:
            # for versions without transitions passed
            return None

    @cached
    def stop_date(self):
        rql = 'Any MIN(D) WHERE E in_state S, WI wf_info_for E,'\
              'WI to_state S, S name IN ("ready", "published"),'\
              'WI creation_date D, E eid %(x)s'
        rset = self._cw.execute(rql, {'x': self.eid}, 'x')
        if rset and rset[0][0]:
            return todate(rset[0][0])
        return None

    def after_deletion_path(self):
        """return (path, parameters) which should be used as redirect
        information when this entity is being deleted
        """
        return self.project.rest_path(), {}

    # ICalendarViews interface ################################################

    def matching_dates(self, begin, end):
        """return prevision or publication date according to state"""
        if self.in_state[0].name == 'published':
            if self.publication_date:
                return [self.publication_date]
        elif self.prevision_date:
            return [self.prevision_date]
        return []

    # ITimetableViews ################

    @property
    def start(self):
        return self.start_date() or self.starting_date or date.today()

    @property
    def stop(self):
        return self.stop_date() or self.prevision_date or date.today()

    # IMileStone interface ####################################################
    parent_type = 'Project'

    def finished(self):
        return self.state in self.FINISHED_STATES

    def in_progress(self):
        return self.state in self.PROGRESSING_STATES

    def initial_prevision_date(self):
        return self.prevision_date

    def completion_date(self):
        return self.publication_date or self.prevision_date

    def get_main_task(self):
        return self.project


    # IPrevNext interface #####################################################

    def previous_entity(self):
        found = False
        for version in self.project.reverse_version_of:
            if found:
                return version
            if version is self:
                found = True

    def next_entity(self):
        return self.next_version(states=None)

    def parent(self):
        return self.project



"""tracker specific hooks

:organization: Logilab
:copyright: 2006-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from itertools import chain

from logilab.common.decorators import classproperty
from logilab.common.compat import any

from cubicweb import ValidationError
from cubicweb.utils import transitive_closure_of
from cubicweb.schema import META_RTYPES
from cubicweb.selectors import implements
from cubicweb.server import hook
from cubicweb.server.hook import (Hook, PropagateSubjectRelationHook,
                                  PropagateSubjectRelationAddHook,
                                  PropagateSubjectRelationDelHook,
                                  match_rtype, match_rtype_sets)
from cubicweb.hooks import notification

# automatization hooks #########################################################

class VersionStatusChangeHook(Hook):
    """when a ticket is done, automatically set its version'state to 'dev' if
      necessary
    """
    __regid__ = 'version_status_change_hook'
    __select__ = Hook.__select__ & hook.match_rtype('in_state',)
    events = ('after_add_relation',)
    ticket_states_start_version = set(('in-progress',))

    def __call__(self):
        entity = self._cw.entity_from_eid(self.eidfrom)
        if entity.e_schema != 'Ticket':
            return
        if entity.state in self.ticket_states_start_version \
               and entity.in_version():
            version = entity.in_version()
            if any(tr for tr in version.possible_transitions()
                   if tr.name == 'start development'):
                version.fire_transition('start development')


class TicketDoneInChangeHook(Hook):
    """when a ticket is attached to a version and it's identical to another one,
    attach the other one as well
    """
    __regid__ = 'ticket_done_in_change'
    __select__ = Hook.__select__ & hook.match_rtype('done_in',)
    events = ('after_add_relation',)

    def __call__(self):
        entity = self._cw.entity_from_eid(self.eidfrom)
        execute = entity._cw.execute
        for identic in entity.identical_to:
            iversion = identic.in_version()
            iveid = iversion and iversion.eid
            if iveid != self.eidto:
                try:
                    execute('SET X done_in V WHERE X eid %(x)s, V eid %(v)s',
                            {'x': identic.eid, 'v': self.eidto}, 'x')
                except:
                    self.exception("can't synchronize version")


class TicketStatusChangeHook(Hook):
    """when a ticket status change and it's identical to another one, change the
    state of the other one as well
    """
    __regid__ = 'ticket_status_change_hook'
    __select__ = Hook.__select__ & implements('TrInfo')
    events = ('after_add_entity',)

    def __call__(self):
        trinfo = self.entity
        forentity = trinfo.for_entity
        if forentity.e_schema != 'Ticket' or not forentity.identical_to:
            return
        synchronized = self._cw.transaction_data.setdefault(
            'ticket_wf_synchronized', set())
        if forentity.eid in synchronized:
            return
        synchronized.add(forentity.eid)
        pstate = trinfo.previous_state
        tr = trinfo.transition
        execute = trinfo._cw.execute
        for identic in forentity.identical_to:
            if identic.eid in synchronized:
                continue
            if identic.current_state and identic.current_state.eid == pstate.eid:
                try:
                    identic.fire_transition(tr.name,
                                            trinfo.comment, trinfo.comment_format)
                except:
                    self.exception("can't synchronize identical ticket's state",
                                   exc_info=True)

# verification hooks ###########################################################

# XXX postpone to an operation, else we may cheat to by-pass this check (prove
# it in a test first!)
class CheckVersionNameOfAProject(Hook):
    """check that the a version of the same project with the same num doesn't
    already exist
    """
    __regid__ = 'check_version_name_of_a_project'
    __select__ = Hook.__select__ & hook.match_rtype('version_of',)
    events = ('before_add_relation',)

    def __call__(self):
        entity = self._cw.entity_from_eid(self.eidfrom)
        project = self._cw.entity_from_eid(self.eidto)
        rset = self._cw.execute(
            'Any X WHERE X num %(num)s, X version_of P, P eid %(p)s',
            {'num': entity.num, 'p': project.eid})
        if rset and (len(rset)>1 or rset[0][0] != entity.eid):
            msg = self._cw._(u'%(vnum)s release number already exists for the project %(prj)s') % {
                'vnum': entity.num, 'prj': project.name}
            raise ValidationError(entity.eid, {"num": msg})


class CheckProjectCyclesHook(Hook):
    __regid__ = 'check_project_cycles'
    __select__ = Hook.__select__ & hook.match_rtype('subproject_of',)
    events = ('before_add_relation',)

    def __call__(self):
        # detect cycles
        proj = self._cw.entity_from_eid(self.eidfrom)
        for child in transitive_closure_of(proj, 'reverse_subproject_of'):
            if self.eidto == child.eid:
                msg = self._cw._('you cannot link these entities, they would form a cycle')
                raise ValidationError(self.eidto, {'child' : msg})


# notification hooks ###########################################################

class BeforeUpdateVersion(notification.EntityUpdateHook):
    __regid__ = 'before_update_version'
    __select__ = Hook.__select__ & implements('Version',)
    skip_attrs = META_RTYPES | set(('description', 'description_format', 'num'))

class BeforeUpdateTicket(notification.EntityUpdateHook):
    __regid__ = 'before_update_ticket'
    __select__ = Hook.__select__ & implements('Ticket',)
    skip_attrs = META_RTYPES | set(('done_in', 'concerns', 'description_format'))

class BeforeInVersionChangeHook(Hook):
    __regid__ = 'before_in_version_change'
    __select__ = Hook.__select__ & hook.match_rtype('done_in',)
    events = ('before_add_relation',)

    def __call__(self):
        if self.eidfrom in self._cw.transaction_data.get('neweids', ()):
            return # entity is being created
        changes = self._cw.transaction_data.setdefault('changes', {})
        thisentitychanges = changes.setdefault(self.eidfrom, set())
        oldrset = self._cw.execute("Any VN WHERE V num VN, T done_in V, T eid %(eid)s",
                                  {'eid': self.eidfrom})
        oldversion = oldrset and oldrset[0][0] or None
        newrset = self._cw.execute("Any VN WHERE V num VN, V eid %(eid)s",
                                  {'eid': self.eidto})
        newversion = newrset[0][0]
        if oldversion != newversion:
            thisentitychanges.add(('done_in', oldversion, newversion))
            notification.EntityUpdatedNotificationOp(self._cw)


# require_permission propagation hooks #########################################

# relations where the "main" entity is the subject
S_RELS = set()
# relations where the "main" entity is the object
O_RELS = set(('concerns', 'version_of', 'subproject_of'))
# not necessary on:
#
# * "secondary" relations: todo_by, done_in, appeared_in, depends_on, uses
# * no propagation needed: wf_info_for
#
# XXX: see_also


class AddEntitySecurityPropagationHook(PropagateSubjectRelationHook):
    """propagate permissions when new entity are added"""
    __regid__ = 'add_entity_security_propagation'
    __select__ = PropagateSubjectRelationHook.__select__ & match_rtype_sets(S_RELS, O_RELS)
    main_rtype = 'require_permission'
    subject_relations = S_RELS
    object_relations = O_RELS


class AddPermissionSecurityPropagationHook(PropagateSubjectRelationAddHook):
    """propagate permissions when new entity are added"""
    __regid__ = 'add_permission_security_propagation'
    __select__ = PropagateSubjectRelationAddHook.__select__ & hook.match_rtype('require_permission',)
    subject_relations = S_RELS
    object_relations = O_RELS

class DelPermissionSecurityPropagationHook(PropagateSubjectRelationDelHook):
    __regid__ = 'del_permission_security_propagation'
    __select__ = PropagateSubjectRelationDelHook.__select__ & hook.match_rtype('require_permission',)
    subject_relations = S_RELS
    object_relations = O_RELS


# has_group_permission propagation hooks #######################################

class AddGroupPermissionSecurityPropagationHook(Hook):
    """propagate on group users when a permission is granted to a group"""
    __regid__ = 'add_group_permission_security_propagation'
    __select__ = Hook.__select__ & hook.match_rtype('require_group',)
    events = ('after_add_relation',)
    rql = ('SET U has_group_permission P WHERE U in_group G, P eid %(p)s, G eid %(g)s,'
           'NOT U has_group_permission P')

    def __call__(self):
        if self._cw.describe(self.eidfrom)[0] != 'CWPermission':
            return
        self._cw.unsafe_execute(self.rql, {'p': self.eidfrom, 'g': self.eidto},)


class DelGroupPermissionSecurityPropagationHook(AddGroupPermissionSecurityPropagationHook):
    """propagate on group users when a permission is removed to a group"""
    __regid__ = 'del_group_permission_security_propagation'
    events = ('after_delete_relation',)
    rql = ('DELETE U has_group_permission P WHERE U in_group G, P eid %(p)s, G eid %(g)s,'
           'NOT EXISTS(U in_group G2, P require_group G2)')


class AddInGroupSecurityPropagationHook(Hook):
    """propagate group permission to users when a permission is granted to a group"""
    __regid__ = 'add_in_group_permission_security_propagation'
    __select__ = Hook.__select__ & hook.match_rtype('in_group',)
    events = ('after_add_relation',)
    rql = ('SET U has_group_permission P WHERE U eid %(u)s, P require_group G, '
           'G eid %(g)s, NOT U has_group_permission P')

    def __call__(self):
        self._cw.unsafe_execute(self.rql, {'u': self.eidfrom, 'g': self.eidto})


class DelInGroupSecurityPropagationHook(AddInGroupSecurityPropagationHook):
    """propagate on existing entities when a permission is deleted"""
    __regid__ = 'del_in_group_permission_security_propagation'
    events = ('after_delete_relation',)
    rql = ('DELETE U has_group_permission P WHERE U eid %(u)s, P require_group G, '
           'G eid %(g)s, NOT EXISTS(U in_group G2, P require_group G2)')



"""views for Version entities

:organization: Logilab
:copyright: 2006-2010 LOGILAB S.A. (Paris, FRANCE), license is LGPL.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from logilab.mtconverter import xml_escape

from cubicweb.selectors import (objectify_selector, implements, score_entity,
                                match_search_state, one_line_rset)
from cubicweb.view import StartupView
from cubicweb import tags
from cubicweb.web import uicfg, component, action
from cubicweb.web.views import primary, baseviews, iprogress, xmlrss

_pvs = uicfg.primaryview_section

_pvs.tag_subject_of(('Version', 'version_of', 'Project'), 'hidden')
_pvs.tag_object_of(('Version', 'version_of', 'Project'), 'hidden')
_pvs.tag_subject_of(('Version', 'depends_on', 'Version'), 'hidden')
_pvs.tag_object_of(('Version', 'depends_on', 'Version'), 'hidden')
_pvs.tag_subject_of(('Version', 'todo_by', 'CWUser'), 'hidden')
for attr in ('num', 'description', 'starting_date', 'prevision_date'):
    _pvs.tag_attribute(('Version', attr), 'hidden')

# primary view and tabs ########################################################


class VersionPrimaryView(primary.PrimaryView):
    __select__ = implements('Version')

    def render_entity_title(self, entity):
        self._cw.add_css('cubes.tracker.css')
        self.w(u'<h1>') #% entity.progress_class())
        self.w('%s %s <span class="state">[%s]</span>' % (
            xml_escape(entity.project.name), xml_escape(entity.num),
            xml_escape(self._cw._(entity.state))))
        self.w(u'</h1>\n')

    def render_entity_attributes(self, entity):
        if entity.description:
            self.w(u'<div><b>%s</b> %s</div>' % (
                self._cw._('focus for this release'),
                entity.view('reledit', rtype='description')))

    def _ticket_table(self, title, ticketrql, nbcol, divid):
        params = {'fname': 'view', 'rql': ticketrql, 'title': title,
                  'vid': 'table', 'fallbackvid': 'null', 'subvid': 'incontext',
                  'displaycols': range(nbcol), 'displayfilter': 1, 'displayactions': 1,
                  'divid': divid,
                  }
        self.w(u'<div class="dynamicFragment" cubicweb:loadurl="%s"></div>'
               % xml_escape(self._cw.build_url('json', **params)))

    def render_entity_relations(self, entity):
        w = self.w; req = self._cw
        req.add_js('cubicweb.ajax.js')
        if entity.conflicts:
            w(u"<div class='entityDescr'><b>%s</b>:<ul>"
              % display_name(self._cw, 'conflicts', context='Version'))
            vid = len(entity.conflicts) > 1 and 'list' or 'outofcontext'
            self.wview(vid, entity.related('conflicts'))
            w(u'</ul></div>')
        # Tickets in version
        req.html_headers.define_var('LOADING_MSG', _('Loading'))
        self._ticket_table(self._cw._('Ticket_plural'), entity.tickets_rql(),
                           entity.tickets_rql_nb_displayed_cols, 'tickets'),
        self._ticket_table(self._cw._('Defects detected on this version'),
                           entity.defects_rql(),
                           entity.defects_rql_nb_displayed_cols, 'defects')


# pluggable sections ###########################################################

class VersionInfoVComponent(component.EntityVComponent):
    """display version information table in the context of the project"""
    __regid__ = 'versioninfo'
    __select__ = component.EntityVComponent.__select__ & implements('Version')
    context = 'navcontenttop'
    order = 10

    def cell_call(self, row, col, view=None):
        self.w(u'<div class="section">')
        version = self.cw_rset.get_entity(row, col)
        view = self._cw.vreg['views'].select('progress_table_view', self._cw,
                                         rset=version.as_rset())
        columns = list(view.columns)
        for col in ('project', 'milestone'):
            try:
                columns.remove(col)
            except ValueError:
                self.warning('could not remove %s from columns' % col)
        view.render(w=self.w, columns=columns)
        self.w(u'</div>')


# secondary views ##############################################################

class VersionTextView(baseviews.TextView):
    __select__ = implements('Version')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(entity.num)
        if entity.in_state:
            self.w(u' [%s]' % self._cw._(entity.state))


class VersionIncontextView(baseviews.InContextView):
    __select__ = implements('Version')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(tags.a(entity.num, href=entity.absolute_url()))


class VersionOutOfContextView(baseviews.OutOfContextView):
    __select__ = implements('Version')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        if entity.version_of:
            project = entity.version_of[0]
            self.w(tags.a(project.name, href=project.absolute_url()))
        self.w(u'&nbsp;-&nbsp;')
        self.w(u'<a href="%s">' % xml_escape(entity.absolute_url()))
        self.wdata(entity.num)
        if entity.in_state:
            self.wdata(u' [%s]' % self._cw._(entity.state))
        self.w(u'</a>')


# other views ##################################################################

class VersionProgressTableView(iprogress.ProgressTableView):
    __select__ = implements('Version')

    title = _('version progression')

    columns = (_('project'), _('milestone'), _('state'), _('planned_start'),
               _('planned_delivery'), _('depends_on'), _('todo_by'))

    def build_depends_on_cell(self, entity):
        vrset = entity.depends_on_rset()
        if vrset: # may be None
            vid = len(vrset) > 1 and 'list' or 'outofcontext'
            return self._cw.view(vid, vrset, 'null')
        return u''

    def build_planned_start_cell(self, entity):
        """``starting_date`` column cell renderer"""
        if entity.starting_date:
            return self._cw.format_date(entity.starting_date)
        return u''

    def build_planned_delivery_cell(self, entity):
        """``initial_prevision_date`` column cell renderer"""
        if entity.finished():
            return self._cw.format_date(entity.completion_date())
        return self._cw.format_date(entity.initial_prevision_date())


class VersionsInfoView(StartupView):
    """display versions in state ready or development, or marked as prioritary.
    """
    __regid__ = 'versionsinfo'
    title = _('All current versions')

    def call(self, sort_col=None):
        rql = ('Any X,P,N,PN ORDERBY PN, version_sort_value(N) '
               'WHERE X num N, X version_of P, P name PN, '
               'EXISTS(X in_state S, S name IN ("dev", "ready")) ')
        rset = self._cw.execute(rql)
        self.wview('progress_table_view', rset, 'null')
        url = self._cw.build_url(rql='Any P,X ORDERBY PN, version_sort_value(N) '
                             'WHERE X num N, X version_of P, P name PN')
        self.w(u'<a href="%s">%s</a>\n'
               % (xml_escape(url), self._cw._('view all versions')))


class RssItemVersionView(xmlrss.RSSItemView):
    __select__ = implements('Version')

    def cell_call(self, row, col):
        entity = self.cw_rset.complete_entity(row, col)
        self.w(u'<item>\n')
        self.w(u'<guid isPermaLink="true">%s</guid>\n'
               % xml_escape(entity.absolute_url()))
        self.render_title_link(entity)
        self._marker('description', entity.dc_description(format='text/html'))
        self._marker('dc:date', entity.dc_date(self.date_format))
        self.render_user_in_charge(entity)
        self.w(u'</item>\n')

    def render_user_in_charge(self, entity):
        if entity.todo_by:
            for user in entity.todo_by:
                self._marker('dc:creator', user.name())


# actions ######################################################################

class VersionAddTicketAction(action.LinkToEntityAction):
    __regid__ = 'addticket'
    __select__ = (action.LinkToEntityAction.__select__ & implements('Version')
                  & score_entity(lambda x: x.state not in x.FINISHED_STATES))

    title = _('add Ticket done_in Version object')
    target_etype = 'Ticket'
    rtype = 'done_in'
    role = 'object'

    def url(self):
        baseurl = super(VersionAddTicketAction, self).url()
        entity = self.cw_rset.get_entity(0, 0)
        linkto = 'concerns:%s:subject' % (entity.version_of[0].eid)
        return '%s&__linkto=%s' % (baseurl, self._cw.url_quote(linkto))

class VersionSubmitBugAction(VersionAddTicketAction):
    __regid__ = 'submitbug'
    __select__ = (action.LinkToEntityAction.__select__ & implements('Version')
                  & score_entity(lambda x: x.state in x.PUBLISHED_STATES))

    title = _('add Ticket appeared_in Version object')
    rtype = 'appeared_in'
    category = 'mainactions'

_abaa = uicfg.actionbox_appearsin_addmenu
_abaa.tag_object_of(('*', 'done_in', 'Version'), False)
_abaa.tag_object_of(('*', 'appeared_in', 'Version'), False)


# register messages generated for the form title until catalog generation is fixed
_('creating Ticket (Ticket done_in Version %(linkto)s)')
_('creating Ticket (Ticket appeared_in Version %(linkto)s)')



"""tracker post creation script, set application's workflows

:organization: Logilab
:copyright: 2003-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from cubes.tracker.schemaperms import xperm

# Ticket workflow
twf = add_workflow(_('default ticket workflow'), 'Ticket')

created    = twf.add_state(_('created'), initial=True)
inprogress = twf.add_state(_('in-progress'))
closed     = twf.add_state(_('closed'))

twf.add_transition(_('start'), created, inprogress,
                   ('managers', 'staff'), xperm('developer'))
twf.add_transition(_('close'), (created, inprogress,), closed,
                   ('managers', 'staff'), xperm('developer'))

# Version workflow
vwf = add_workflow(_('default version workflow'), 'Version')

planned   = vwf.add_state(_('planned'), initial=True)
dev       = vwf.add_state(_('dev'))
ready     = vwf.add_state(_('ready'))
published = vwf.add_state(_('published'))

vwf.add_transition(_('start development'), planned, dev,
                   ('managers', 'staff'), xperm('developer'))
vwf.add_transition(_('ready'), dev, ready,
                   ('managers', 'staff'), xperm('developer'))
vwf.add_transition(_('publish'), (dev, ready), published,
                   ('managers', 'staff'), xperm('developer'))
vwf.add_transition(_('stop development'), dev, planned,
                   ('managers', 'staff'), xperm('developer'))



# pylint: disable-msg=W0622
"""cubicweb-tracker application packaging information

Copyright (c) 2003-2010 LOGILAB S.A. (Paris, FRANCE).
http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
modname = 'tracker'
distname = 'cubicweb-tracker'

numversion = (1, 5, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'

author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'

short_desc = 'issue tracker component for the CubicWeb framework'
long_desc = '''Summary
-------

The `tracker` cube provides an issue tracker with projects, tickets and versions
to group tickets. You can defines roadmap for a project by planning future
versions.

A link to subscribe to rss feed for published versions of project is
located on each project toolbar.

Usage
-----

- ProjectRoadmapView, displays all version

- ProjectStatsView, displays the number of tickets for each status
  ('done', 'validation pending', and so on)
'''

from os import listdir as _listdir
from os.path import join, isdir

web = 'http://www.cubicweb.org/project/%s' % distname
classifiers = [
          'Environment :: Web Environment',
          'Framework :: CubicWeb',
          'Programming Language :: Python',
          'Programming Language :: JavaScript',
          'Topic :: Software Development :: Bug Tracking',
]

pyversions = ['2.4']

#from cubicweb.devtools.pkginfo import get_distutils_datafiles
CUBES_DIR = join('share', 'cubicweb', 'cubes')
THIS_CUBE_DIR = join(CUBES_DIR, 'tracker')

def listdir(dirpath):
    return [join(dirpath, fname) for fname in _listdir(dirpath)
            if fname[0] != '.' and not fname.endswith('.pyc')
            and not fname.endswith('~')
            and not isdir(join(dirpath, fname))]

from glob import glob
try:
    data_files = [
        # common files
        [THIS_CUBE_DIR, [fname for fname in glob('*.py') if fname != 'setup.py']],
        ]
    # check for possible extended cube layout
    for dirname in ('entities', 'views', 'sobjects', 'hooks', 'schema', 'data', 'i18n', 'migration', 'wdoc'):
        if isdir(dirname):
            data_files.append([join(THIS_CUBE_DIR, dirname), listdir(dirname)])
    # Note: here, you'll need to add subdirectories if you want
    # them to be included in the debian package
except OSError:
    # we are in an installed directory
    pass


__depends_cubes__ = {}
__depends__ = {'cubicweb': '>= 3.6.0'}
__use__ = tuple(__depends_cubes__)

